#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from .Operator import Operator, LinearOperator, ScaledOperator, SumOperator,\
 CompositionOperator, AdjointOperator
from .BlockOperator import BlockOperator
from .SparseFiniteDifferenceOperator import SparseFiniteDifferenceOperator
from .FiniteDifferenceOperator import FiniteDifferenceOperator
from .GradientOperator import GradientOperator
from .SymmetrisedGradientOperator import SymmetrisedGradientOperator
from .IdentityOperator import IdentityOperator
from .ZeroOperator import ZeroOperator
from .MatrixOperator import MatrixOperator
from .DiagonalOperator import DiagonalOperator
from .MaskOperator import MaskOperator
from .ChannelwiseOperator import ChannelwiseOperator
from .BlurringOperator import BlurringOperator
from .ProjectionMap import ProjectionMap
